/*-
 * Copyright (c) 2009-2010 UPLEX - Nils Goroll Systemoptimierung
 * Partly based upon code Copyright (c) 2008 Linpro AS
 * All rights reserved.
 *
 * Authors: Nils Goroll <nils.goroll@uplex.de>
 *	    Geoffrey Simmons <geoff.simmons@uplex.de>
 *
 * Partly based upon code by Petter Knudsen <petter@linpro.no>
 *	   
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 *
 * VSLP*) Director - Varnish StateLess Persistence Director
 *
 * The purpose of this director is to provide stateless persistence (or
 * persistency if you prefer) for backend requests based upon this simple
 * concept:
 *
 * - Generate a load balancing key, which will be used to select the backend
 *
 *   The key values should be as uniformly distributed as possible.
 *
 *   For all requests which need to hit the same backend server, the same
 *   key must be generated.
 *
 *   For strings, a hash function is used to generate the key.
 *
 * - Select the preferred backend server as (key mod nservers), nservers being
 *   the number of backend servers for this director.
 *
 * - (Only) If the preferred backend server becomes unavailable, an alternative
 *   backend server is selected using the same method from the set of healhy
 *   servers.
 *
 * This technique allowes to implement persistence without keeping any state,
 * and, in particular, without the need to synchronize state between nodes of a
 * cluster of Varnish servers.
 *
 * Still, all members of a cluster will, if otherwise configured equally,
 * implement the same persistence, in other words, requests sharing a common
 * persistence criterium will be balanced onto the same backend server no matter
 * which Varnish server handles the request.
 *
 * The drawbacks are:
 *
 * - the distribution of requests depends on the number of requests per key and
 *   the uniformity of the distribution of key values. In short, this technique
 *   will generally lead to less good load balancing compared to stateful
 *   techniques.
 *
 * - When a backend server becomes unavailable, every persistence technique has
 *   to reselect a new backend server, but this technique will also switch back
 *   to the preferred server once it becomes healthy again, so the persistence
 *   is generally less stable compared to stateful techniques.
 *
 * - Persistence will generally be lost when the number of hosts configured for
 *   the director is changed (so the configuration should be kept stable).
 *
 * *) The name VSLP was chosen to accompany VTLA as another VETR (Varnish
 *    Exception To the Rule) ;-) http://varnish-cache.org/wiki/VTLA
 */

#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "shmlog.h"
#include "cache.h"
#include "cache_backend.h"
#include "vrt.h"


/*--------------------------------------------------------------------*/

struct vdi_vslp_host {
	struct backend			*backend;
	struct vdi_vslp_host		*alternative;
};

#if 0
/* MOVED -> cache_backend.h */
struct vdi_vslp {
	unsigned		magic;
#define VDI_VSLP_MAGIC		0xcd0d15f3 /* crc32 */
	struct director		dir;
	struct vdi_vslp_host	*hosts;
	unsigned		nhosts;

	/* default digest, if not set in per session data */
	enum vbe_digest_e	digest;
};
#endif

static inline int
vdi_vslp_rampup(struct sess *sp, struct vdi_vslp *vs, uint32_t key, 
		unsigned int chosen, double t_since_recovery) {
	int	i;
	double	r;
	
	/* 
	 * The chosen host is within the rampup interval Decide whether to use
	 * it or pick another host Probabilty of using it increases linearly
	 * with time since recovery
	 */

	VSL(SLT_Director, sp->id, "host %s in rampup",
	    vs->hosts[chosen].backend->vcl_name);

	r = random() / 2147483648.0;	/* 2^31 */
	assert(r >= 0.0 && r < 1.0);

	if (r * vs->rampup > t_since_recovery) {
		/*
		 * Pick another host
		 * Use the recovery times of all other hosts to estimate which
		 * ones were healthy at the time of chosen's recovery
		 */

		int nhealthies = 0;

		for (i = 0; i < vs->nhosts; i++)
			if (vs->hosts[i].backend->recovered == 0.0 ||
			    (vs->hosts[i].backend->recovered < 
			     vs->hosts[chosen].backend->recovered))
				vs->healthies[nhealthies++] = i;
		/*
		 * Assuming that key hasn't changed (although it could have, if
		 * the VCL has changed).
		 */

		if (nhealthies) {
			chosen = vs->healthies[key % nhealthies];
			VSL(SLT_Director, sp->id, "choosing host %s instead",
			    vs->hosts[chosen].backend->vcl_name);
		}
	}

	return (chosen);
}


static struct vbe_conn *
vdi_vslp_try(struct sess *sp, struct vdi_vslp *vs, unsigned int i, struct director_sess *ds, const char *type_str)
{
	struct backend		*backend;
	struct vbe_conn		*vbe;

	backend = vs->hosts[i].backend;
	if (backend->healthy && (vbe = VBE_GetVbe(sp, backend))) {
		VSL(SLT_Director, sp->id, "name %s vcl_name %s "
		    "%s %s healthy for %s",
		    vs->dir.name,
		    vs->dir.vcl_name,
		    type_str,
		    backend->vcl_name,
		    sp->http->hd[HTTP_HDR_URL].b);

		if (! ds->first_tried)
			ds->first_tried = backend;

		ds->last_tried = backend;

		return (vbe);
	}
	return (NULL);
}

#define VDI_VSLP_TRY(sp,vs,i,ds,s)			\
	if ((vbe = vdi_vslp_try(sp, vs, i, ds, s)))	\
		return (vbe)

static struct vbe_conn *
vdi_vslp_getfd(struct sess *sp)
{
	unsigned int		i, chosen;
	unsigned int		nhealthy = 0;

	uint32_t		key;

	struct vdi_vslp		*vs;
	struct vbe_conn		*vbe;
	struct director_sess	*ds;

	double r;

	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CHECK_OBJ_NOTNULL(sp->director, DIRECTOR_MAGIC);
	CAST_OBJ_NOTNULL(vs, sp->director->priv, VDI_VSLP_MAGIC);

	ds = VDI_get_session(sp);

#if NEW_CODE
	/* if we are retrying, go to to each alternative server in turn */
	if (ds->first_tried) {
		struct backend	*backend = ds->last_tried;

		AN(backend);
		while ((backend = backend ... )) {}
	}
#endif

	if ((key = ds->persistence_key)) {
		key--;
	} else {
		/* No valid key. Fall back to hash of URL */
		key = crc32_l(sp->http->hd[HTTP_HDR_URL].b, 
		    strlen(sp->http->hd[HTTP_HDR_URL].b));
		ds->persistence_key = key;
	}

	/*
	 * 1st step: chose server from all configured servers, so
	 *	     each document is fetched from a preferred server
	 *	     whenever this one is healthy (and the list of
	 *	     servers does not change)
	 */
	
	chosen = key % vs->nhosts;

	/*
	 * If the primary backend is healthy, we might sill use
	 * another one for two reasons:
	 *
	 * - the primary backend has just become healthy and still
	 *   is in the rampup period or
	 *
	 * - we have an alternative server probability set and we
	 *   want to warm up that server. The alternative server
	 *   is always the next server in the ring, so it does
	 *   not change when the health state changes.
	 *
	 */
	
	if (vs->hosts[chosen].backend->healthy) {
		double t_since_recovery = TIM_mono() -
		    vs->hosts[chosen].backend->recovered;

		DSL(0x02, SLT_Debug, sp->id,
		    "%s: t_since_recovery = %f, rampup = %f",
		    vs->hosts[chosen].backend->vcl_name,
		    t_since_recovery,
		    vs->rampup);
		
		if ((vs->rampup > 0.0) && (t_since_recovery < vs->rampup)) {
			unsigned int rampup_chosen =
			    vdi_vslp_rampup(sp, vs, key, chosen,
			                    t_since_recovery);
			VDI_VSLP_TRY(sp, vs, rampup_chosen, ds,
			    (rampup_chosen == chosen) ?
			    "preferred rampup" :
			    "alternative rampup");
		} else if (vs->altsrv_p > 0.0) {
			/* from cache_dir_random.c */
			r = random() / 2147483648.0;	/* 2^31 */
			assert(r >= 0.0 && r < 1.0);
			if (r < vs->altsrv_p) {
				VDI_VSLP_TRY(sp, vs,
				    (chosen + 1) % vs->nhosts,
				    ds, "altsrv_p");
			}
		}
		VDI_VSLP_TRY(sp, vs, chosen, ds, "preferred");
	}

	/* 2nd step: try alternate */
	if (vs->altsrv_p > 0.0)
		VDI_VSLP_TRY(sp, vs,
		    (chosen + 1) % vs->nhosts,
		    ds, "alternate");
	
	/*
	 * 3rd step: If the preferred backend is unavailable, we
	 *	     choose an alternative "sticky" backend from
	 *	     those alive which will stay the same as long
	 *	     as the number of healthy backend nodes remains
	 *	     unchanged
	 */

	for (i = 0; i < vs->nhosts; i++) {
		if (vs->hosts[i].backend->healthy)
			nhealthy++;
	}

	if (nhealthy == 0) {
		VSL(SLT_Director, sp->id, "name %s vcl_name %s none healthy",
		    vs->dir.name,
		    vs->dir.vcl_name);
		
		return (NULL);
	}
	
	chosen = key % nhealthy;

	/* 
	 * chosen is now an index into the virtual list of healthy
	 * backends - so we have to iterate once more
	 */
	
	for (i = 0; i < vs->nhosts; i++) {
		if (!vs->hosts[i].backend->healthy)
			continue;

		if (chosen-- <= 0) {
			VDI_VSLP_TRY(sp, vs, i, ds, "last");
		}
	}

	VSL(SLT_Director, sp->id, "name %s vcl_name %s"
	    " NO host found for %s",
	    vs->dir.name,
	    vs->dir.vcl_name,
	    sp->http->hd[HTTP_HDR_URL].b);
	return (NULL);
}

static unsigned
vdi_vslp_healthy(const struct sess *sp)
{
	struct vdi_vslp *vs;
	int i;

	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CHECK_OBJ_NOTNULL(sp->director, DIRECTOR_MAGIC);
	CAST_OBJ_NOTNULL(vs, sp->director->priv, VDI_VSLP_MAGIC);

	for (i = 0; i < vs->nhosts; i++) {
		if (vs->hosts[i].backend->healthy)
			return 1;
	}
	return 0;
}

/*lint -e{818} not const-able */
static void
vdi_vslp_fini(struct director *d)
{
	int i;
	struct vdi_vslp *vs;
	struct vdi_vslp_host *vh;

	CHECK_OBJ_NOTNULL(d, DIRECTOR_MAGIC);
	CAST_OBJ_NOTNULL(vs, d->priv, VDI_VSLP_MAGIC);

	vh = vs->hosts;
	for (i = 0; i < vs->nhosts; i++, vh++)
		VBE_DropRef(vh->backend);
	free(vs->hosts);
	free(vs->dir.vcl_name);
	free(vs->healthies);
	vs->dir.magic = 0;
	FREE_OBJ(vs);
}

void
VRT_init_dir_vslp(struct cli *cli, struct director **bp,
    const struct vrt_dir_vslp *t)
{
	struct vdi_vslp *vs;
	const struct vrt_dir_vslp_entry *te;
	struct vdi_vslp_host *vh;
	int i;

	(void)cli;

	ALLOC_OBJ(vs, VDI_VSLP_MAGIC);
	XXXAN(vs);
	vs->hosts = calloc(sizeof *vh, t->nmember);
	XXXAN(vs->hosts);

	vs->dir.magic = DIRECTOR_MAGIC;
	vs->dir.priv = vs;
	vs->dir.name = "vslp";
	REPLACE(vs->dir.vcl_name, t->name);
	vs->dir.getfd = vdi_vslp_getfd;
	vs->dir.fini = vdi_vslp_fini;
	vs->dir.healthy = vdi_vslp_healthy;

	vh = vs->hosts;
	te = t->members;
	for (i = 0; i < t->nmember; i++, vh++, te++)
		vh->backend = VBE_AddBackend(cli, te->host);
	vs->nhosts = t->nmember;

	vs->altsrv_p = t->altsrv_p;
	vs->rampup = t->rampup;
	vs->healthies = calloc(sizeof(int), vs->nhosts);

	*bp = &vs->dir;
}
